'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function input(node, value) {
    const listener = () => { var _a; return (_a = value.set) === null || _a === void 0 ? void 0 : _a.call(value, node.value); };
    if (value.set)
        node.addEventListener('input', listener);
    return {
        update() {
            node.value = value.get();
        },
        destroy() {
            if (value.set)
                node.removeEventListener('input', listener);
        },
    };
}
function entries(object) {
    return Object.getOwnPropertyNames(object).map(key => [key, object[key]]);
}
function event(node, value) {
    const listeners = new Map(entries(value.get()).map(([key, listener]) => {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
        node.addEventListener(key, listener);
        return [key, listener];
    }));
    return {
        update() {
            const events = value.get();
            for (const [name, listener] of listeners) {
                if (events[name] !== listener) {
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
                    node.removeEventListener(name, listener);
                    listeners.set(name, events[name]);
                }
            }
        },
        destroy() {
            for (const [name, listener] of listeners.entries())
                // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
                node.removeEventListener(name, listener);
        }
    };
}
function click(node, value) {
    let listener = value.get();
    node.addEventListener('click', listener);
    return {
        update() {
            node.removeEventListener('click', listener);
            listener = value.get();
            node.addEventListener('click', listener);
        },
        destroy() {
            node.removeEventListener('click', listener);
        }
    };
}

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

class EventEmitter {
    constructor() {
        Object.defineProperty(this, "listeners", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
    }
    on(event, listener) {
        let listeners;
        let index = 0;
        if (this.listeners.has(event))
            index = (listeners = this.listeners.get(event)).push(listener) - 1;
        else
            this.listeners.set(event, listeners = [listener]);
        return {
            revoke: () => {
                // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
                delete listeners[index];
            }
        };
    }
    trigger(event, ...args) {
        if (!this.listeners.has(event))
            return;
        this.listeners.get(event).forEach(listener => listener(...args));
    }
}

class Modifiers {
    constructor() {
        Object.defineProperty(this, "modifiers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new WeakMap()
        });
    }
    set(target, key, modifier, value) {
        let map;
        if (this.modifiers.has(target))
            map = this.modifiers.get(target);
        else
            // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
            this.modifiers.set(target, map = new Map());
        if (map.has(key))
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            map.get(key)[modifier] = value;
        else
            map.set(key, { [modifier]: value });
    }
    get(target, key, modifier) {
        var _a, _b;
        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
        return (_b = (_a = this.modifiers.get(Object.getPrototypeOf(target))) === null || _a === void 0 ? void 0 : _a.get(key)) === null || _b === void 0 ? void 0 : _b[modifier];
    }
    has(target, key, modifier) {
        return this.get(target, key, modifier) !== undefined;
    }
}

class Scheduler {
    constructor() {
        Object.defineProperty(this, "queue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Set()
        });
        Object.defineProperty(this, "microtask", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: Promise.resolve()
        });
        Object.defineProperty(this, "flushing", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
    }
    next(cb) {
        void this.microtask.then(cb);
    }
    flush() {
        for (const task of this.queue) {
            task();
            this.queue.delete(task);
        }
    }
    enqueue(task) {
        this.queue.add(task);
        if (!this.flushing) {
            this.next(() => {
                this.flushing = false;
                this.flush();
            });
        }
    }
}

/* eslint-disable */
function isEqualRef(x, y) {
    return !x.find((key, i) => key !== y[i]);
}
const modifiers = new Modifiers();
function assertNotModifier(target, key, modifier, message) {
    if (modifiers.has(target, key, modifier))
        throw new Error(message);
}
function assertNotIgnored(target, key, message = `Property ${String(key)} is ignored.`) {
    assertNotModifier(target, key, 'ignore', message);
}
class Model {
    constructor() {
        /** @internal */
        Object.defineProperty(this, "$__events", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new EventEmitter()
        });
        /** @internal */
        Object.defineProperty(this, "$__scheduler", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Scheduler()
        });
        const extend = (object, parent) => {
            const ref = (key) => [...parent !== null && parent !== void 0 ? parent : [], key];
            let refNext = false;
            const getters = new WeakMap();
            const computed = (target, key, receiver, descriptior) => {
                var _a;
                if (getters.has(descriptior.get))
                    return getters.get(descriptior.get)();
                let value;
                let dirty = true;
                let dispose;
                let debounceTimeout;
                let observers = [];
                const extenders = (_a = modifiers.get(target, key, 'extend')) !== null && _a !== void 0 ? _a : [];
                const computed = () => {
                    if (dirty) {
                        let dependencies = [];
                        const listener = this.$__events.on('get', ref => dependencies.push(ref));
                        const next = Reflect.get(target, key, receiver);
                        listener.revoke();
                        dirty = false;
                        if (next != value) {
                            value = next;
                            for (const extender of extenders) {
                                if (extender.compute)
                                    value = extender.compute(value);
                            }
                            dispose === null || dispose === void 0 ? void 0 : dispose();
                            const listeners = dependencies.map(dependency => this.$__observe(dependency, () => __awaiter(this, void 0, void 0, function* () {
                                dirty = true;
                                for (const extender of extenders) {
                                    if (extender.recompute && !(yield extender.recompute())) {
                                        dirty = false;
                                        return;
                                    }
                                    if (extender.notify && !extender.notify(value))
                                        return;
                                }
                                const notify = () => {
                                    this.$__events.trigger('set', ref(key));
                                    for (const listener of observers)
                                        listener();
                                };
                                if (modifiers.has(target, key, 'throttle')) {
                                    const timeout = modifiers.get(target, key, 'throttle');
                                    setTimeout(notify, timeout);
                                }
                                else if (modifiers.has(target, key, 'debounce')) {
                                    const timeout = modifiers.get(target, key, 'debounce');
                                    if (debounceTimeout)
                                        clearTimeout(debounceTimeout);
                                    debounceTimeout = setTimeout(notify, timeout);
                                }
                                else {
                                    this.$schedule(notify);
                                }
                            })));
                            dispose = () => listeners.map(l => l.revoke());
                        }
                    }
                    return value;
                };
                const observe = (listener) => {
                    const index = observers.push(listener) - 1;
                    return {
                        revoke() {
                            delete observers[index];
                        }
                    };
                };
                for (const extender of extenders) {
                    if (extender.initialize) {
                        extender.initialize({
                            get: computed,
                            observe,
                            dispose() {
                                throw new Error('Cannot dispose of computed property.');
                            },
                        });
                    }
                }
                getters.set(descriptior.get, computed);
                return computed();
            };
            return new Proxy(object, {
                get: (target, key, receiver) => {
                    if (!parent && key === this.$reference.name)
                        refNext = true;
                    if ((!parent && typeof key === 'string' && key.startsWith('$'))
                        || modifiers.get(target, key, 'ignore'))
                        return Reflect.get(target, key, receiver);
                    if (refNext) {
                        refNext = false;
                        return ref(key);
                    }
                    const get = () => {
                        const descriptior = Object.getOwnPropertyDescriptor(Object.getPrototypeOf(target), key);
                        if (descriptior === null || descriptior === void 0 ? void 0 : descriptior.get)
                            return computed(target, key, receiver, descriptior);
                        return Reflect.get(target, key, receiver);
                    };
                    const value = get();
                    this.$__events.trigger('get', ref(key));
                    if (typeof value === 'object' && value !== null)
                        return extend(value, ref(key));
                    return value;
                },
                set: (target, key, value, receiver) => {
                    const succeeded = Reflect.set(target, key, value, receiver);
                    this.$schedule(() => {
                        this.$__events.trigger('set', ref(key));
                    });
                    return succeeded;
                }
            });
        };
        const extended = extend(this);
        for (const key in this.constructor.prototype) {
            if (Object.prototype.hasOwnProperty.call(this.constructor.prototype, key)) {
                if (modifiers.get(this.constructor.prototype, key, 'effect'))
                    Reflect.get(extended, key);
            }
        }
        return extended;
    }
    static ignore(target, key) {
        assertNotIgnored(target, key, `Property ${String(key)} is already ignored.`);
        modifiers.set(target, key, 'ignore', true);
    }
    static effect(target, key) {
        assertNotIgnored(target, key);
        modifiers.set(target, key, 'effect', true);
    }
    static throttle(delay) {
        return (target, key) => {
            assertNotIgnored(target, key);
            assertNotModifier(target, key, 'throttle', `Property ${String(key)} is already throttled.`);
            assertNotModifier(target, key, 'debounce', `Cannot throttle property ${String(key)}. Property is already debounced.`);
            modifiers.set(target, key, 'throttle', delay);
        };
    }
    static debounce(delay) {
        return (target, key) => {
            assertNotIgnored(target, key);
            assertNotModifier(target, key, 'debounce', `Property ${String(key)} is already debounced.`);
            assertNotModifier(target, key, 'throttle', `Cannot debounce property ${String(key)}. Property is already throttled.`);
            modifiers.set(target, key, 'debounce', delay);
        };
    }
    static readonly(target, key) {
        assertNotModifier(target, key, 'readonly', `Property ${String(key)} is already readonly.`);
        modifiers.set(target, key, 'readonly', true);
    }
    static extend(extender) {
        return (target, key) => {
            if (modifiers.has(target, key, 'extend'))
                modifiers.get(target, key, 'extend').push(extender);
            else
                modifiers.set(target, key, 'extend', [extender]);
        };
    }
    /** @internal */
    $__observe(property, listener) {
        return this.$__events.on('set', (ref) => {
            if (isEqualRef(property, ref))
                listener();
        });
    }
    /** @internal */
    $__observeDependencies(callback, listener) {
        let dependencies = [];
        const dependencyListener = this.$__events.on('get', (ref) => dependencies.push(ref));
        callback();
        dependencyListener.revoke();
        const listeners = dependencies.map(dependency => this.$__observe(dependency, () => {
            listener();
        }));
        return {
            revoke() {
                for (const listener of listeners)
                    listener.revoke();
            },
        };
    }
    /**
     * Creates a reference to the property.
     *
     * @example
     * ```typescript
     * this.$reference(this.myProp)
     * ```
     *
     * @param target The property in the model to reference
     */
    $reference(target) {
        const ref = target;
        let subscriptions = [];
        let disposed = false;
        const assertNotDisposed = () => {
            if (disposed)
                throw new Error('Reference has been disposed.');
        };
        return {
            get: () => {
                assertNotDisposed();
                const keys = [...ref];
                let target = this[keys.shift()];
                for (const key of keys) {
                    if (key in target)
                        target = target[key];
                    else
                        return undefined;
                }
                return target;
            },
            set: (value) => {
                assertNotDisposed();
                const keys = [...ref];
                const key = keys.pop();
                let target = keys.length ? this[keys.shift()] : this;
                for (const key of keys) {
                    if (key in target)
                        target = target[key];
                    else
                        return false;
                }
                target[key] = value;
                return true;
            },
            observe: (listener) => {
                assertNotDisposed();
                const subscription = this.$__observe(ref, listener);
                subscriptions.push(subscription);
                return subscription;
            },
            dispose: () => {
                assertNotDisposed();
                disposed = true;
                for (const subscription of subscriptions)
                    subscription.revoke();
            },
        };
    }
    /**
     * Creates an computed _value_ which recomputes when a dependency in the current model changes.
     *
     * @example
     * ```typescript
     * // This will recompute when the value of `x` or `y` changes.
     * this.$computed(() => {
     * 	return this.x + this.y
     * })
     * ```
     */
    $computed(get) {
        let subscriptions = [];
        let value;
        let dirty = true;
        let dispose;
        let disposed = false;
        const assertNotDisposed = () => {
            if (disposed)
                throw new Error('Reference has been disposed.');
        };
        return {
            get: () => {
                assertNotDisposed();
                if (dirty) {
                    let dependencies = [];
                    const listener = this.$__events.on('get', ref => dependencies.push(ref));
                    const next = get();
                    listener.revoke();
                    dirty = false;
                    if (next != value) {
                        value = next;
                        dispose === null || dispose === void 0 ? void 0 : dispose();
                        const listeners = dependencies.map(dependency => this.$__observe(dependency, () => {
                            dirty = true;
                            this.$__scheduler.enqueue(() => {
                                subscriptions.forEach(listener => listener());
                            });
                        }));
                        dispose = () => listeners.map(l => l.revoke());
                    }
                }
                return value;
            },
            observe: (listener) => {
                assertNotDisposed();
                const index = subscriptions.push(listener) - 1;
                return {
                    revoke() {
                        delete subscriptions[index];
                    }
                };
            },
            dispose: () => {
                assertNotDisposed();
                disposed = true;
                dispose === null || dispose === void 0 ? void 0 : dispose();
            },
        };
    }
    $schedule(task) {
        this.$__scheduler.enqueue(task);
    }
    $tick() {
        this.$__scheduler.flush();
    }
}

exports.Model = Model;
exports.click = click;
exports.event = event;
exports.input = input;
